
#include "tools/utils.h"

#include "clustering.h"
#include "timing.h"

#include <vector>
#include "cudpp.h"

namespace CUDA
{
  Timing timing;

  #include "helpers.cuh"
  #include "cuda_kernel.cuh"

  const size_t BLOCKSIZE = 256;
 
  template<class T>
  T divup(T a, T b) { return (a + b - 1)/b; }

  uint *d_activeset, *d_activeset_offset, *d_leaves, *d_leaves_offset, *d_triangle_offset;
  CUDPPHandle theCudpp;
  CUDPPHandle scanplan;

  void prepareData(CUDAStorages* memory)
  {
     setOctreeBoundaries(memory->boundaries);
     setLevelOffsets(memory->h_levelOffset.get());

     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(octree, &memory->d_octree_data, sizeof(OctreeEntry*)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(vertexCoherence, &memory->d_octree_vertex_coherence, sizeof(const unsigned int*)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(vertexIdOut, &memory->d_vertex_lookup, sizeof(unsigned int *)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(vertices, &memory->vertexcount, sizeof(unsigned int)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(maxdepth, &memory->maxdepth, sizeof(unsigned int)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(voxelIds, &memory->d_voxel_ids, sizeof(const uint3*)));
     CUDA_CHECKED_CALL(cudaMemcpyToSymbol(triangleCoherence, &memory->d_octree_triangle_coherence, sizeof(const unsigned int*)));

     //data arrays
     CUDA_CHECKED_CALL(cudaMalloc(&d_activeset, sizeof(uint)*memory->voxelcount));
     CUDA_CHECKED_CALL(cudaMalloc(&d_activeset_offset, sizeof(uint)*(memory->voxelcount+1)));
     CUDA_CHECKED_CALL(cudaMalloc(&d_leaves, sizeof(uint)*(memory->voxelcount)));
     CUDA_CHECKED_CALL(cudaMalloc(&d_leaves_offset, sizeof(uint)*(memory->voxelcount+1)));
     CUDA_CHECKED_CALL(cudaMalloc(&d_triangle_offset, sizeof(uint)*(memory->trianglecount)));

     cudppCreate(&theCudpp);
     CUDPPConfiguration config;
     config.op = CUDPP_ADD;
     config.datatype = CUDPP_UINT;
     config.algorithm = CUDPP_SCAN;
     config.options = CUDPP_OPTION_FORWARD | CUDPP_OPTION_EXCLUSIVE;
     size_t scanSize = std::max<size_t>(memory->voxelcount+1,memory->trianglecount);
     CUDPPResult res = cudppPlan(theCudpp, &scanplan, config, scanSize, 1, 0);
     if (CUDPP_SUCCESS != res)
       std::cout << "Error creating CUDPPPlan\n";

  }


  void cleanUp()
  {
     CUDA_CHECKED_CALL(cudaFree(d_activeset));
     CUDA_CHECKED_CALL(cudaFree(d_activeset_offset));
     CUDA_CHECKED_CALL(cudaFree(d_leaves));
     CUDA_CHECKED_CALL(cudaFree(d_leaves_offset));
     CUDA_CHECKED_CALL(cudaFree(d_triangle_offset));
     cudppDestroyPlan(scanplan);
     cudppDestroy(theCudpp);
  }

  size_t runPrefixSum(uint* d_input, uint num)
  {
    uint numEnd0, numEnd1;
    CUDA_CHECKED_CALL(cudaMemcpy(&numEnd0, d_input+(num-1), sizeof(uint), cudaMemcpyDeviceToHost));
    CUDPPResult res = cudppScan(scanplan, d_input, d_input, num);
    if (CUDPP_SUCCESS != res)
        std::cout << "Error running scan\n";
    CUDA_CHECKED_CALL(cudaMemcpy(&numEnd1, d_input+(num-1), sizeof(uint), cudaMemcpyDeviceToHost));
    return numEnd0 + numEnd1;
  }
  
  template<bool  DEPTHERROR>
  size_t buildMesh(CUDAStorages* memory, float threshold);

  size_t buildMesh(CUDAStorages* memory, float threshold, bool error_depth)
  {
    if(error_depth)
      return buildMesh<true>(memory,  threshold);
    else
      return buildMesh<false>(memory, threshold);
  }

  template<bool  DEPTHERROR>
  size_t buildMesh(CUDAStorages* memory, float threshold)
  {
    CUDA::GL::mapped_buffer<unsigned int> d_indices(memory->cuda_index_buffer);
    CUDA_CHECKED_CALL(cudaMemcpyToSymbol(triOutput, &d_indices.get(), sizeof(uint*)));

    size_t numnodes = 0;
    size_t numleafs = 0;
    size_t activenodes = 1;
    size_t cdepth = 0;
    
    timing.begin();

    uint null = 0;
    CUDA_CHECKED_CALL(cudaMemcpy(d_activeset, &null, sizeof(uint), cudaMemcpyHostToDevice));
    CUDA_CHECKED_CALL(cudaMemcpy(d_triangle_offset, &null, sizeof(uint), cudaMemcpyHostToDevice));

    while(activenodes > 0)
    {
      checkLevel<DEPTHERROR><<<divup(activenodes,BLOCKSIZE), BLOCKSIZE>>>(d_activeset + numnodes, d_activeset_offset, d_leaves_offset, cdepth, activenodes, threshold);
      size_t newNodes = runPrefixSum(d_activeset_offset, activenodes);
      size_t newLeafs = runPrefixSum(d_leaves_offset, activenodes);
      writeLevel<DEPTHERROR><<<divup(activenodes,BLOCKSIZE), BLOCKSIZE>>>(d_activeset + numnodes, d_activeset_offset, d_leaves_offset, 
                                                              d_activeset + numnodes + activenodes,  d_triangle_offset + numnodes + activenodes,  d_leaves + numleafs,
                                                              cdepth, activenodes, threshold);
      //write vertices
      if(newLeafs != 0)
        handleLeafs<<<divup(newLeafs,BLOCKSIZE), BLOCKSIZE>>>(d_leaves + numleafs,  newLeafs);

      numnodes += activenodes;
      numleafs += newLeafs;
      activenodes = newNodes;
      ++cdepth;
    }

    
    //write triangles
    size_t tris = runPrefixSum(d_triangle_offset, numnodes);
    handleNodes<<<divup(numnodes,BLOCKSIZE), BLOCKSIZE>>>(d_activeset,  d_triangle_offset, numnodes);
    cudaDeviceSynchronize();
    double dt = timing.end();
    std::cout << "CUDA: " << dt*1000 << " (" << numnodes << " nodes)" <<  std::endl;
    return 3*tris;
  }

};