
#include "gpuscheduling.cu"

#include "image.h"
#include "renderingdata.cuh"
#include "corefunctions.cuh"

#include "timing.h"

struct Params
{
   unsigned int* p_randseeds;
   rgbcolor* p_imagedata;
}

struct TraceWp : public PayloadedWorkpackage<Params>
{
  uint2 pix;
  TraceWp(uint3 _pix, uint threads, Params* _params) : 
    pix(_pix.x, _pix.y), 
    PayloadedWorkpackage<Params>(threads, _params)
  { }
};


class TracingProcedure : public Procedure
{
public:
  //macro will activate the third tier for this procedure
  ACTIVATE_THIRD_TIER

  __device__ static void execute(volatile Workpackage* workpackage)
  {

    //define additional payload directly in execute
    PAYLOAD_BEGIN
    vector3 contrib;
    vector3 orig;
    rgbcolor color;
    vector3 dir;
    unsigned int randseed;
    uint recursion;
    PAYLOAD_END

    volatile TraceWp* mywp = static_cast<volatile TraceWp*>(workpackage);
    uint2 pixel = mywp->pix;

    //init data structures
	  randseed =  getSeed(mywp->params->p_randseeds, pixel, __threadId());

    contrib = make_float3(1.0f);
    orig = make_float3(0.0f);
    color = make_float3(0.0f);

    dir.x = ( (float)pixel.x / (float)c_width ) - 0.5f + device_random(randseed)/(float)c_width;
	  dir.y = ( ( (float)pixel.y / (float)c_height ) - 0.5f ) * ( (float)c_height/(float)c_width) + device_random(randseed)/(float)c_height;
	  dir.z = 1;
	  dir = vec_normalize( dir );

    recursion = 0;

    //scheduler is activated every 10th iteration of this loop
    SCHEDULEDWHILE(true, 10)
    {
      if( !( ( recursion < c_maxrecursions ) && ( ( contrib.x > MIN_CONTRIB ) || ( contrib.y  > MIN_CONTRIB ) || ( contrib.z > MIN_CONTRIB ) ) ) )
      {
        rgbcolor* outcolor = mywp->params->p_imagedata + (pixel.y*c_width + pixel.x);
        atomicAdd(&outcolor->x, color.x);
        atomicAdd(&outcolor->y, color.y);
        atomicAdd(&outcolor->z, color.z);
        SCHEDULEENDTHREAD;
      }

      
      float3 d, n;
      float t;
      const object* obj = intersectObjects(c_objects, c_numObjects, d, n, t, orig, dir);

      if(obj) 
        hitObject(*obj, t, n, orig, dir, contrib, color, randseed);
      else
		  {
        color += c_background * contrib;
		    contrib.x = contrib.y = contrib.z = 0;
		  }
      recursion++;
	  }
    SCHEDULEDWHILEEND
  }  
};

double renderScheduled(int device, const Scene& scene)
{
  Scheduler scheduler(device);

  setupData(scene);
  initImage();
  initRand();

  Params params;
  params.p_imagedata = d_imagedata;
  params.p_randseeds = d_randseeds;

  auto tracingEvent = scheduler.createEvent<TraceWp, TracingProcedure, Params>();

  double t0 = Timing::gettime();

  dim imageDims(scene.imgWidth(), scene.imgHeight());
  tracingEvent->fire(DefaultCreatorParams(imageDims, tsamples, params));
  TracingEvent->join();

  double t1 = Timing::gettime();
  return t1-t0;
}


#endif